/*
  sample.c
  David Rowe 13 January 2009

  User mode side of Mesh Potato Serial I/O test program.  Reads a tx
  files of chars while writing a rx sample of chars.

  Credits: Serial port init code lifted from ser.c,
  http://www.captain.at
*/

/*
  Copyright (C) 2009 David Rowe
	    (C) 2004-5 Captain http://www.captain.at
 
  All rights reserved.
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2, as
  published by the Free Software Foundation.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */

#define MP_BUF_SZ     160
#define FS            8000
#define MAX_STR       256

int initport(int fd) {
    struct termios options;

    // Get the current options for the port...

    tcgetattr(fd, &options);

    // Set the baud rates to 115200...

    cfsetispeed(&options, B115200);
    cfsetospeed(&options, B115200);

    // Enable the receiver and set local mode...

    options.c_cflag |= (CLOCAL | CREAD);

    options.c_cflag &= ~PARENB;
    options.c_cflag &= ~CSTOPB;
    options.c_cflag &= ~CSIZE;
    options.c_cflag |=  CS8;
    options.c_cflag &= ~ECHO;
    options.c_cflag &= ~ECHONL;

    if (options.c_cflag & ECHO)
	printf("ECHO\n");
    if (options.c_cflag & ECHONL)
	printf("ECHONL\n");

    // Set the new options for the port...

    cfmakeraw(&options);
    tcsetattr(fd, TCSANOW, &options);

    return 1;
}

int main(int argc, char *argv[]) {
    int    fd, fdserial, len, i, frames;
    char   txbuf[MP_BUF_SZ];
    char   rxbuf[MP_BUF_SZ];
    FILE  *ftx, *frx;
    float  secs;

    if (argc < 4) {
	printf("usage: %s TxFileName RxFileName length(secs)\n", argv[0]);
	printf("e.g. : ./sample /tmp/tx.raw /tmp/rx.raw 3\n\n");
	printf("TxFileName and RxFileName should be on a ram disk (e.g. \\tmp), "
	       "as the router serial flash is too slow and will cause audio "
	       "break up.  If you use a non-existant name for TxFileName audio "
	       "will only be recorded (not played).  The audio format is 8 bit "
	       "ulaw at 8000 samples/sec.  scp to your Linux x86 box and replay "
               " with:\n $ play -f U -r 8000 -s b rx.raw -v 5 \n");
	exit(0);
    }

    ftx = fopen(argv[1],"rb");
    if (ftx == NULL) {
	printf("Can't open tx sample file: %s\n", argv[2]);
	printf("Continuing with no tx....\n");
	memset(txbuf, MP_BUF_SZ, 0);
    }

    frx = fopen(argv[2],"wb");
    if (frx == NULL) {
	printf("Can't open rx sample file: %s\n", argv[3]);
	exit(1);
    }

    secs = atof(argv[3]);
    if ((secs < 0.0) || (secs > 100.0)) {
	printf("Invalid secs %f, must be between 0 and 100\n", secs);
	exit(1);
    }
    frames = (secs*FS)/MP_BUF_SZ;

    /* open serial port so we can set speed */

    fdserial = open("/dev/ttyS0", O_RDWR | O_NOCTTY | O_NDELAY);
    if (fdserial == -1) {
	perror("open_port: Unable to open /dev/ttyS0");
	return 1;
    } else {
	fcntl(fdserial, F_SETFL, 0);
    }
	
    initport(fdserial);

    fd = open("/dev/8250mp", O_RDWR);
    if( fd == -1) {
	printf("/dev/8250mp open error...\n");
	close(fdserial);
	exit(0);
    }
	
    printf("sampling.....\n");
    for(i=0; i<frames; i++) {
	len = read(fd, rxbuf, MP_BUF_SZ);
	if( len == -1 ) {
	    printf("read error...\n");
	    exit(1);
	}
	fwrite(rxbuf, sizeof(char), MP_BUF_SZ, frx);

	if (ftx != NULL)
	    fread(txbuf, sizeof(char), MP_BUF_SZ, ftx);
	len = write(fd, txbuf, MP_BUF_SZ);
	if( len == -1 ) {
	    printf("write error...\n");
	    exit(1);
	}
    }

    close(fdserial);
    close(fd);
    if (ftx != NULL)
	fclose(ftx);
    fclose(frx);

    return 0;
}
